<?php

/**
 *
 * @author Peter Csubak (chuby)
 * @copyright Peter Csubak (chuby)
 */
interface XmlSerializable2{
	
	public function beforeSerialize($shemaName);
	public function afterSerialize($shemaName, $xml);
	public function beforeDeserialize($schemaName, $xml);
	public function afterDeserialize($schemaName, $xml);
	public function manualDeserialize($schema, $xml, $fieldName);
	public function manualSerialize($shema, $fieldName);

}
?>